/****************************************************************
/*  Program Name: Balloon.c
/*	Version: 2002c.0.2
/*	Date: 8 June 2002
/*
/*	This source was written by Zack Clobes, W0ZC for use
/*	with the Project: Traveler APRS tracking board.  It is
/*	free for all non-commerical use.  You may contact me
/*	at zack@clobes.net.
/*
/*	Periodic updates and/or modifications will be made
/*	available at www.rckara.org/.
/*
/*	While I trust this code with my own equipment, I make
/*	no warranty that it will function correctly in any given
/*	environment and will not be held resposible for data or
/*	property loss due to the failure of this code or 
/*	related hardware.
/***************************************************************/

#include <16f876.h>

#fuses  HS,WDT,NOPROTECT,NOLVP
#use delay(clock=20000000, RESTART_WDT)
#use rs232(baud=4800, rcv=PIN_C5, ERRORS, INVERT)

//constants to setup the analog port
#bit ADFM_BIT = 0x9f.7
#define AD_LEFT_JUSTIFY 0
#define AD_RIGHT_JUSTIFY 1


//Pin Assignments
//#define IN_ADC_OAT 0			// AN0 - Aux 1 - this pin is in-op
#define IN_ADC_AUX2 1			// AN1 - Aux 2
#define IN_ADC_IAT 2			// AN2 - Also Aux 3 if temp sensor removed
#define IN_ADC_BATTERY 4		// AN4


#define IN_DIG_AUX1 48			// RB0/Int
#define IN_DIG_AUX2 49			// RB1
#define IN_DIG_AUX3 50			// RB2
#define OUT_DIG_CAMERA 51		// RB3
#define IN_DIG_AUX5 52			// RB4
#define IN_DIG_AUX6 53			// RB5
#define IN_DIG_AUX7 54			// RB6
#define IN_DIG_AUX8 55			// RB7
#define IN_DIG_TNC_RX_DETECT 56	// RC0
#define IN_DIG_TNC_DATA_RX 57	// RC1



#define OUT_DIG_TNC_M0 60		// RC4
#define OUT_DIG_TNC_M1 59		// RC3
#define OUT_DIG_TNC_DATA_TX 58		// RC2
#define OUT_DIG_TNC_PTT 44		// RA4


#define MAX_LEN_TEMP 85 //Sets the maximum buffer size that can be transmitted in one string
#define TIME_BETWEEN_XMITS 2000	//number of milliseconds between individual transmissions
#define FREQUENCY_OF_STATUS_MESSAGE 7	//Number of transmission cycles between sending status/telemetry key
#define SEC_DELAY_BETWEEN_CYCLES 45	//One second delays between  cycles

//global variables-----------------------------------------------------
char szTemp[MAX_LEN_TEMP];	// szTemp to hold strings of data to send to TNC
int iCyclesToNextParamsStatus;	//number of 1 minute cycles between status and parameter key xmissions
long lCurrentAlt;	//contains current altitude in meters
long lMaxAlt;		//contains burst altitude in meters
long lLastPictureTakenAlt;	//altitude when last picture was taken - used to determine next picture altitude
long lNextPictureAlt;	//the altitude to take the next picture
long lCurrentTime;	//The current GPS time
long lMaxAltTime;	//The time of the burst
boolean bBurst;		//flag to indicate that we have burst

//function prototypes--------------------------------------------------
#include "AX25.C"

void initController(void);
void getGPRMC(void);
void getGPGGA(void);
void getCurrentTimeAltitude(void);
void checkForBurst(void);
long getNextPictureAltitude(void);
void xmitLine(void);
void xmitTelemetry(void);
long getAnalog(int iAnalogPort);
signed int convertToCelsius(long lValue);
signed int floatToSignedInt(float fValue);
long convertTomV(long lValue);
void takePicture(void);

//---------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////////
void main(void) {
	int i;
	
	
	//The calibrate function is used to calibrate the timing cycle used for 1200 baud generation.
	// When this function is un-commented, the controller will go into a endless loop of sending
	// High-low-high-low... to the modem.
	//calibrate();		//Used for debugging and calibrating the AX25 routines

	//determine how we are booting: Watchdog timeout, normal boot, or other
	switch (restart_cause()) {
		case WDT_TIMEOUT: {
			initController();
			strcpy(szTemp, ">Watchdog timeout ->POSSIBLE GPS FAILURE<-  Controller re-initializing...");
			xmitLine();
			restart_wdt();
			// if there was a watchdog reset, then transmit the telemetry data
			delay_ms(TIME_BETWEEN_XMITS);
			restart_wdt();
			xmitTelemetry();	//transmitting telemetry at least gives a RF source to 
						// track - use zero for time so we aren't relying on GPS
			restart_wdt();
			break;
		}
		case NORMAL_POWER_UP: {
			initController();

			strcpy(szTemp, ">RCKARA Balloon Controller Initializing...");
			xmitLine();
			restart_wdt();
			break;
		}
		default: {
			initController();
			strcpy(szTemp, ">UNKNOWN STARTUP CONDITION OCCURRED  Controller re-initializing...");
			xmitLine();
			restart_wdt();
			delay_ms(TIME_BETWEEN_XMITS);
			restart_wdt();
			xmitTelemetry();
			restart_wdt();
			break;
		}
	}




	// go into infinate loop to send packets
	do {

// ------------------------ TRANSMISSION BATCH #1 ------------------------
		for (i=0;i<SEC_DELAY_BETWEEN_CYCLES;i++) {
			restart_wdt();
			getCurrentTimeAltitude();
			restart_wdt();
			checkForBurst();	//check for descending altitude (i.e. balloon burst)
			restart_wdt();
			
			lNextPictureAlt = getNextPictureAltitude();	//store the altitude to take the next pic
			
			//determine if we're rising or falling
			if (bBurst == FALSE) {
				//we're rising
				if ((lCurrentAlt >= lNextPictureAlt) && (lNextPictureAlt != 0)) {
					takePicture();
					lLastPictureTakenAlt = lCurrentAlt;
				}
			} else {
				//we're falling
				if ((lCurrentAlt <= lNextPictureAlt) && (lNextPictureAlt != 0)) {
					takePicture();
					lLastPictureTakenAlt = lCurrentAlt;
				}
			}
			restart_wdt();
		}
		//transmit the RMC & GGA sentence (Altitude included) and telemetry for 2002c
		getGPRMC();
		xmitLine();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();


		getGPGGA();
		restart_wdt();
		xmitLine();
		restart_wdt();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();

		getCurrentTimeAltitude();	//gets and stores the current time/altitude
		xmitTelemetry();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();



		//Occasionally transmit a parameters key to the telemetry data as well as
		// a status message from the controller.
		// The frequency of these messages is determined by the define FREQUENCY_OF_STATUS_MESSAGE
		if (iCyclesToNextParamsStatus == 0) {
			strcpy(szTemp, "{{PTime:hhmm,Burst:m,IAT:C,PriBatt:mV -- Mode 2");
			xmitLine();
			delay_ms(TIME_BETWEEN_XMITS);

			//I usually use this next message to indicate that this is a balloon
			// launch and give a URL for more information.
			strcpy(szTemp, ">Project: Traveler Balloon - www.rckara.org");
			xmitLine();
			iCyclesToNextParamsStatus = FREQUENCY_OF_STATUS_MESSAGE;	//reset the counter
		} else iCyclesToNextParamsStatus--;	//decr the counter if we didn't hit zero already


		

// ------------------------ TRANSMISSION BATCH #2 ------------------------
		//now wait another 30 seconds before transmitting RMC, telemetry...
		for (i=0;i<SEC_DELAY_BETWEEN_CYCLES;i++) {
			//wait for 27 altitude reports (30 seconds)
			restart_wdt();
			getCurrentTimeAltitude();
			restart_wdt();
			checkForBurst();	//check for descending altitude (i.e. balloon burst)
			restart_wdt();
			
			lNextPictureAlt = getNextPictureAltitude();	//store the altitude to take the next pic
			
			//determine if we're rising or falling
			if (bBurst == FALSE) {
				//we're rising
				if ((lCurrentAlt >= lNextPictureAlt) && (lNextPictureAlt != 0)) {
					takePicture();
					lLastPictureTakenAlt = lCurrentAlt;
				}
			} else {
				//we're falling
				if ((lCurrentAlt <= lNextPictureAlt) && (lNextPictureAlt != 0)) {
					takePicture();
					lLastPictureTakenAlt = lCurrentAlt;
				}
			}
			restart_wdt();
		}

		//beginning of the main transmission group
		getGPRMC();
		xmitLine();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();

		getGPGGA();
		xmitLine();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();

		getCurrentTimeAltitude();	//gets and stores the current time/altitude
		xmitTelemetry();
		delay_ms(TIME_BETWEEN_XMITS);
		restart_wdt();

	} while (TRUE);
}
/////////////////////////////////////////////////////////////////////////////////
void initController(void) {
//This function initializes the modem and any global variables/flags
	//starts the WDT every 2.3 secs
	setup_counters(RTCC_INTERNAL, WDT_2304MS);
	restart_wdt();


	initTNC();	//calls function in AX25.C to initialize the pins to the modem

	//init the secondary output pins
	output_low(OUT_DIG_CAMERA);


	delay_ms(10000);	//wait for the transmitter to fire up
	restart_wdt();

	//configure the analogs ports
	setup_adc_ports(A_ANALOG_RA3_REF);	//must use RA3 as a reference voltage since the uP runs at 4.7V
	setup_adc(ADC_CLOCK_DIV_8);
	ADFM_BIT = AD_RIGHT_JUSTIFY; //put the data into the right hand side
	iCyclesToNextParamsStatus = 0;	//Transmit the Status/telemetry key the first time around

	//init the secondary variables (non-critical)
	lCurrentAlt = 0;
	lMaxAlt = 0;
	lCurrentTime = 0;
	lMaxAltTime = 0;
	lLastPictureTakenAlt = 0;
	bBurst = FALSE;
}
/////////////////////////////////////////////////////////////////////////////////
void getGPRMC(void) {
//This function loops through the incoming GPS data until it finds $GPRMC and then it collects 
// any text that follows it
	int i;
	restart_wdt();
start:
	while (szTemp[0] != '$') (szTemp[0] = getc());   //finds the $GPRMC sentence
	
	szTemp[1] = getc();
	if (szTemp[1] != 'G'){
		goto start;
	}
	szTemp[2] = getc();
	if (szTemp[2] != 'P'){
		goto start;
	}
	szTemp[3] = getc();
	if (szTemp[3] != 'R'){
		goto start;
	}
	szTemp[4] = getc();
	if (szTemp[4] != 'M'){
		goto start;
	}
	szTemp[5] = getc();
	if (szTemp[5] != 'C'){
		goto start;
	}

	for (i = 6;i < MAX_LEN_TEMP;i++) {
		szTemp[i] = getc();
		if (szTemp[i] <= 13) {
			//we've found the end of the sentence
			szTemp[i] = 0;
			return;
		}
	}
	szTemp[MAX_LEN_TEMP - 1] = 0;	// if we go all the way out, make sure it's null terminated	
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
void getGPGGA(void) {
//This function loops through the incoming GPS data until it finds $GPGGA and then it collects 
// any text that follows it
	int i;
	restart_wdt();
start:
	while (szTemp[0] != '$') (szTemp[0] = getc());   //finds the $GPGGA sentence

	
	szTemp[1] = getc();
	if (szTemp[1] != 'G'){
		goto start;
	}
	szTemp[2] = getc();
	if (szTemp[2] != 'P'){
		goto start;
	}
	szTemp[3] = getc();
	if (szTemp[3] != 'G'){
		goto start;
	}
	szTemp[4] = getc();
	if (szTemp[4] != 'G'){
		goto start;
	}
	szTemp[5] = getc();
	if (szTemp[5] != 'A'){
		goto start;
	}

	for (i = 6;i < MAX_LEN_TEMP;i++) {
		szTemp[i] = getc();
		if (szTemp[i] <= 13) {
			//we've found the end of the sentence
			szTemp[i] = 0;
			return;
		}
	}
	szTemp[MAX_LEN_TEMP - 1] = 0;	// if we go all the way out, make sure it's null terminated	
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
void getCurrentTimeAltitude(void) {
//This function sets the current time and current altitude variables from the GGA sentence 
	char c, cAlt[5];	
	int i;
	long lDecPlace;
	
	restart_wdt();
start:
	while (szTemp[0] != '$') (szTemp[0] = getc());   //finds the $GPGGA sentence


	c = getc();
	if (c != 'G'){
		goto start;
	}
	c = getc();
	if (c != 'P'){
		goto start;
	}
	c = getc();
	if (c != 'G'){
		goto start;
	}
	c = getc();
	if (c != 'G'){
		goto start;
	}
	c = getc();
	if (c != 'A'){
		goto start;
	}

	c = getc(); 	//get the extra comma
	
	//start getting time
	lCurrentTime = ((long int)(getc()) - 48) * 1000;	//tens digit of hour
	lCurrentTime += ((long int)(getc()) - 48) * 100;	//ones digit of hour
	lCurrentTime += ((long int)(getc()) - 48) * 10;	//tens digit of minutes
	lCurrentTime += ((long int)(getc()) - 48);	//ones digit of minutes


	for (i=0;i<8;i++) {	//wait for 9 commas before continuing
		c = getc();	//must get the next character before entering loop again
		while (c != ',') {
			c = getc();	//loop until a comma is found
		}
	}
	
	//build the altitude string up to the decimal
	i = 0;	// set the counter
	c = getc();
	
	//loop while we're getting strictly numeric values
	while ((c >= '0') && (c <= '9')) {
		cAlt[i] = c;
		i++;
		c = getc();
	}
	
	//loop backward thru the string and reconstruct the long type from ASCII digits
	lDecPlace = 1;	//initially set to the ones digit
	lCurrentAlt = 0;	//init the return variable
	while (i > 0) {
		i--;
		lCurrentAlt += (long int)(cAlt[i] - 48) * lDecPlace;	
		if (lDecPlace < 10000) lDecPlace *= 10;	//get the dec place holder ready for 
							// next dec place only if it won't 
							// get too big for long
	}
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
void checkForBurst(void) {
//This function watches the current and previous altitudes to see when we start falling.
// When we do, it sets the lBurstAlt variable to display the max altitude
	
	//see if we're higher than the last recorded max alt
	if (lCurrentAlt > lMaxAlt) {
		//we have a new max altitude so store it
		lMaxAlt = lCurrentAlt;
		lMaxAltTime = lCurrentTime;	//and the time it occurred
	} else {
		//check to see if we're high enough for a burst (over 1500m),
		// and if we've fallen enough to conclude burst (dropped at least 200m from max alt)
		// and make sure we haven't already burst
		if ((lMaxAlt > 1500) && ((lMaxAlt - 200) > lCurrentAlt) && (!bBurst)) {
			bBurst = TRUE;
			lLastPictureTakenAlt = 65535;	//some big number
		}
	}
}
/////////////////////////////////////////////////////////////////////////////////
long getNextPictureAltitude(void) {
//This function is essentially a lookup table to determine when to take pictures
// under ascending and descending conditions.
	if (bBurst == FALSE) {
		//we're still rising, so check this table
		if (lLastPictureTakenAlt < 610) return 610;
		if (lLastPictureTakenAlt < 910) return 910;
		if (lLastPictureTakenAlt < 1220) return 1220;
		if (lLastPictureTakenAlt < 1520) return 1520;
		if (lLastPictureTakenAlt < 1830) return 1830;
		if (lLastPictureTakenAlt < 2440) return 2440;
		if (lLastPictureTakenAlt < 3050) return 3050;
		if (lLastPictureTakenAlt < 3660) return 3660;
		if (lLastPictureTakenAlt < 4570) return 4570;
		if (lLastPictureTakenAlt < 6100) return 6100;
		if (lLastPictureTakenAlt < 7620) return 7620;
		if (lLastPictureTakenAlt < 9150) return 9150;
		if (lLastPictureTakenAlt < 12190) return 12190;
		if (lLastPictureTakenAlt < 15240) return 15240;
		if (lLastPictureTakenAlt < 18290) return 18290;
		if (lLastPictureTakenAlt < 21340) return 21340;
		if (lLastPictureTakenAlt < 22870) return 22870;
		if (lLastPictureTakenAlt < 24390) return 24390;
		if (lLastPictureTakenAlt < 25910) return 25910;
		if (lLastPictureTakenAlt < 27440) return 27440;
		return 0;	//we're done taking pics at this time
	} else {
		//we're falling so check this table
		if (lLastPictureTakenAlt > 22870) return 22870;
		if (lLastPictureTakenAlt > 19820) return 19820;
		if (lLastPictureTakenAlt > 16770) return 16770;
		if (lLastPictureTakenAlt > 13720) return 13720;
		if (lLastPictureTakenAlt > 10670) return 10670;
		if (lLastPictureTakenAlt > 7620) return 7620;
		if (lLastPictureTakenAlt > 6100) return 6100;
		if (lLastPictureTakenAlt > 5490) return 5490;
		if (lLastPictureTakenAlt > 4880) return 4880;
		if (lLastPictureTakenAlt > 4270) return 4270;
		if (lLastPictureTakenAlt > 3660) return 3660;
		if (lLastPictureTakenAlt > 3050) return 3050;
		if (lLastPictureTakenAlt > 2740) return 2740;
		if (lLastPictureTakenAlt > 2440) return 2440;
		if (lLastPictureTakenAlt > 2130) return 2130;
		if (lLastPictureTakenAlt > 1830) return 1830;
		if (lLastPictureTakenAlt > 1520) return 1520;
		if (lLastPictureTakenAlt > 1220) return 1220;
		if (lLastPictureTakenAlt > 910) return 910;
		if (lLastPictureTakenAlt > 610) return 610;
		return 0;	//we're done taking pics
	}
}
/////////////////////////////////////////////////////////////////////////////////
void xmitLine(void) {
//This function handles all of the packet headers and cleanup needed to send a text string
// with only one line of code
	restart_wdt();
	sendPacketStart();
	restart_wdt();
	sendSZString(szTemp);
	restart_wdt();
	sendCrLf();
	restart_wdt();
	sendPacketEnd();
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
void xmitTelemetry(void) {
//This function collects all of the input data and formulates the telemetry packet.
// PT MODE 2
	long lIAT;
	long lBattery;	//vars to hold raw data (before conversion)
	signed int iIAT;			//vars to hold the temp in deg celsius
		
	restart_wdt();
	lIAT = getAnalog(IN_ADC_IAT);
	iIAT = convertToCelsius(lIAT);

	lBattery = getAnalog(IN_ADC_BATTERY);
	lBattery = convertTomV(lBattery);
	lBattery *= 3.12765957;		//compensate for resistor divider on input
	restart_wdt();

	sendPacketStart();
	strcpy(szTemp, "{{D");	//this is a custom packet type
	sendSZString(szTemp);
	restart_wdt();
	
	writeTime(lCurrentTime);
	sendByte(',');
	restart_wdt();
	
	//only xmit the max alt if we've burst
	if (bBurst) {
		writeLongDec(lMaxAlt, TRUE);
	} else {
		writeLongDec(0, TRUE);
	}
	sendByte(',');
	restart_wdt();
	
	writeSignedIntDec(iIAT, TRUE);
	sendByte(',');
	restart_wdt();

	writeLongDec(lBattery, TRUE);
	sendCrLf();
	sendPacketEnd();
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
long getAnalog(int iAnalogPort) {
//This function get the analog data from the ADCs and returns it as a long
	long lData;

	set_adc_channel(iAnalogPort);
	delay_ms(50);	//you're supposed to delay before reading the analog input
	lData = Read_ADC();
	restart_wdt();
	return lData;
}
/////////////////////////////////////////////////////////////////////////////////
signed int convertToCelsius(long lValue) {
//This function does the math to convert the raw analog input value into degree celsius
	float fTemp;

	//eq. Temp = (lValue - 195) / 2.031746
	lValue -= 195;
	fTemp = (float)lValue;
	fTemp /= 2.031746;
	restart_wdt();
	return (signed int)fTemp;
}
/////////////////////////////////////////////////////////////////////////////////
long convertTomV(long lValue) {
//This function does the math to convert the raw analog input value into millivolts
// It does not compensate for the resistor divider used to measure voltages greater than 5.0V	
	float fTemp;
	fTemp = lValue * 4.98046875;
	restart_wdt();
	return (long int)fTemp;
}
/////////////////////////////////////////////////////////////////////////////////
void takePicture(void) {
	output_high(OUT_DIG_CAMERA);
	restart_wdt();
	delay_ms(400);
	restart_wdt();
	output_low(OUT_DIG_CAMERA);
	restart_wdt();
}
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////